/*----------------------------------------------------------------------------

  NLHDR - Non local high dynamic range image generation

  Copyright (c) 2013 cecilia aguerrebere <caguerrebere@gmail.com>
   
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Affero General Public License for more details.

  This code is partially based on the implementation by A. Buades
  
  A. Buades, B. Coll, and J.-M. Morel. Non-Local Means Denoising. 
  Image Processing On Line, 2011

  ----------------------------------------------------------------------------*/

#include <float.h>  
#include "nlhdr.h"

void   nlhdr  (int iDWin,	    /* Half patch size */
	       int iDBloc,	    /* Half size of research window */
	       float ***fpI,	    /* Input images */
	       float **fpO,	    /* Output image */
	       int iFrames,      /* Number of input frames */
	       int iChannels,    /* Number of input image channels */ 
	       int iWidth,       /* Number of columns of input images */
	       int iHeight,      /* Number of rows of input images */
	       float* tau,       /* Exposure times vector */
	       int ref_t,        /* Reference frame index */
	       float ih,         /* Threshold on patch distances */
	       int* trans_y,     /* Vector of vertical (row) translations of each frame with respect to the reference frame*/
	       int* trans_x,     /* Vector of horizontal (column) translations of each frame with respect to the reference frame*/
	       float muR,        /* Camera parameters: readout noise mean */
	       float sigma2R,    /*                    readout noise variance */
	       float *g,         /*                    gain */
	       float **prnu,     /*                    PRNU factors map */
	       float iVsat)      /*                    saturation threshold */

{

  int totalPix = (iDWin*2+1)*(iDWin*2+1);  /* number of pixels in a patch */
  int iwxh = iWidth * iHeight;             /* length of each channel */
  float distSatPatchs = FLT_MAX;
  int minPix = totalPix / 2;	 /* Minimum number of non saturated pixels required on a
                                    patch to be used for distance computation*/
  int progress = 0;	
  int z,c;
				
  /* Allocate memory for the irradiance and variance maps */  
  double *rad[iFrames][iChannels];		
  double *var[iFrames][iChannels];
  double *radDvar[iFrames][iChannels];
  double *invVar[iFrames][iChannels];
  int *satMap[iFrames][iChannels];
  for (z = 0; z < iFrames; z++) 
    for (c = 0; c < iChannels; c++){
      rad[z][c] = (double*) malloc(iwxh*sizeof(double));
      var[z][c] = (double*) malloc(iwxh*sizeof(double));
      radDvar[z][c] = (double*) malloc(iwxh*sizeof(double));
      invVar[z][c] = (double*) malloc(iwxh*sizeof(double));
      satMap[z][c] = (int*) malloc(iwxh*sizeof(int));
    }
	
  /* Normalize frames */
  for (z = 0; z < iFrames; z++) {
    if ( z != ref_t ) {
      fnormalize_frame( fpI[z], rad[z], var[z], radDvar[z], invVar[z], iWidth, iHeight, prnu, muR, sigma2R, g, tau[z], iChannels );
      fsaturation_maps( fpI[z], satMap[z], iWidth, iHeight, iChannels, iVsat );
    } else {  /* use higher iVsat for the reference frame in order to force no saturated samples. */
      fnormalize_frame( fpI[z], rad[z], var[z], radDvar[z], invVar[z], iWidth, iHeight, prnu, muR, sigma2R, g, tau[z], iChannels );
      fsaturation_maps( fpI[z], satMap[z], iWidth, iHeight, iChannels, iVsat*1000 );      
    }
  }


#pragma omp parallel 
  {		 		
    /* For each pixel on the reference frame... */	 				        
#pragma omp for schedule(dynamic) 
    for(int y=0; y < iHeight ; y++) {      								
      for(int x=0 ; x < iWidth;  x++)
	{															
  
	  /* Check if the central pixel of the current patch is not saturated for all channels. 
	     Otherwise the pixel is not denoised. Denoising differently on the channels can produce
	     color artifacts. */
	  int centerNotSaturated2 = 0;	  	  
	  for (int ii=0; ii < iChannels; ii++) 
	    /* The center pixel of the current patch is not saturated */
	    if ( satMap[ref_t][ii][y*iWidth + x] == 1 ) 
	      centerNotSaturated2++;
	    
				
	  /* If the four channels are not saturated then denoise the pixel */			
	  if ( centerNotSaturated2 == iChannels )  {
		  
	    double fTotalWeight[iChannels];   
	    double denoised_pix[iChannels];   
	    for (int ii=0; ii < iChannels; ii++) {
	      fTotalWeight[ii] = 0.0;				
	      denoised_pix[ii] = 0.0;				
	    }
			  
	    /* Search for similar pixels in all frames (reference frame and the other frames) */	    
	    for (int z = 0; z < iFrames; z++) {
		  
	      /* Reduce the size of the comparison window if near the boundary */
	      int iDWin0 = MIN(iDWin,MIN(iWidth-1-x-trans_x[z],MIN(iHeight-1-y-trans_y[z],MIN(x-trans_x[z],y-trans_y[z]))));				      
					
	      /* Research zone depending on the boundary and the size of the window */
	      int imin=MAX(x-trans_x[z]-iDBloc,iDWin0);
	      int jmin=MAX(y-trans_y[z]-iDBloc,iDWin0);					
	      int imax=MIN(x-trans_x[z]+iDBloc,iWidth-1-iDWin0);
	      int jmax=MIN(y-trans_y[z]+iDBloc,iHeight-1-iDWin0);
		
	      /* Search for similar pixels on the search window */ 		
	      if ( iDWin0 > 0 && jmax > jmin && imax > imin ) 	      
	      
		for(int j=jmin; j <= jmax; j++)
		  for(int i=imin ; i <= imax; i++) 
		    /* Check it is not the current pixel */   
		    if (i!=x || j!=y || z!=ref_t)
		      {
			/* Check if the central pixel of the patch is not saturated for all channels. 
			   Otherwise discard the patch. Denoising differently on the channels can produce
			   color artifacts. */
			int centerNotSaturated = 0;
			for (int ii=0; ii < iChannels; ii++) 
			  /* The center pixel of the current patch is not saturated */
			  if ( satMap[z][ii][j*iWidth + i] == 1 ) {
			    centerNotSaturated++;
			  }
			/* The four channels are not saturated */	
			if ( centerNotSaturated	== iChannels )  {
				
			  /*------------ Compute the weights -----------------------*/
						
			  /* Compute squared difference between patches */
			  int total_pix_used;
			  float fDif = fpatch_dist(rad[ref_t],rad[z],var[ref_t],var[z],satMap[ref_t],satMap[z],x,y,i,j,iDWin0,iChannels,iWidth,iWidth,minPix,distSatPatchs,&total_pix_used);						
								
			  /* Set binary weights: */			
			  /* If the distance fDif is below the threshold 
			   * dist_threshold2 then weight = 1 */
			  /* Otherwise weight = 0 */
			  float total_pix_usedf = (float) total_pix_used;
			  float dist_threshold =  total_pix_usedf * ((1 + tau[ref_t] / tau[z])*(0.5 + ih / sqrt(2*total_pix_usedf)) - 1);
			  int fWeight = (fDif <= dist_threshold) ? 1 : 0;				   			    				
			
			  /*------------ Denoise patch central pixel ---------------*/	
			  /* Denoise the pixel and compute the total weight */
			  if ( fWeight == 1  )		    
			    for (int ii=0; ii < iChannels; ii++) {			    
			      double *patch_center = &radDvar[z][ii][j*iWidth + i];
			      double *patch_centerW = &invVar[z][ii][j*iWidth + i];
			      denoised_pix[ii] += (double)(*patch_center);			      
			      fTotalWeight[ii] += (double)(*patch_centerW);				      
			    }
			}	
		      }
	   
	    } /* end searching for similar pixels in frame z */ 
	    
	    /* Include the central pixel on the denoising */
	    for (int ii=0; ii < iChannels; ii++) {
		  
	      double *patch_center = &radDvar[ref_t][ii][y*iWidth + x];
	      double *patch_centerW = &invVar[ref_t][ii][y*iWidth + x];
	      denoised_pix[ii] += (double)*patch_center;
	      fTotalWeight[ii] += (double)*patch_centerW;

	      /* Fill output image */
	      if ( fTotalWeight[ii] > dTiny  ) 		
	      /* After removing the readout noise offset some pixels may take negative values
	       * and give negative irradiance estimations. That represents the lack of irradiance, 
	       * so set negative irradiance estimations to zero. */	
		fpO[ii][y*iWidth + x] = denoised_pix[ii] / fTotalWeight[ii] > 0 ? (float) (denoised_pix[ii] / fTotalWeight[ii]) : 0.;				    			      
	      else 
		fpO[ii][y*iWidth + x] = rad[ref_t][ii][y*iWidth + x] > 0 ? (float) rad[ref_t][ii][y*iWidth + x] : 0.; 			    			      
	    } 
			
	  } else { /* The central pixel of the patch is saturated*/

	    /* Fill output image */
		    		    
	    /* If the sample is saturated, the radiance value will be noisy since it is 
	       equal to vSat / (prnu * tau * g). Instead of assigning the same constant value to all
	       radiance values of saturated samples, different values are assigned because of the division
	       by prnu. Do the following to avoid that noise. */
	    for (int ii=0; ii < iChannels; ii++) {	        
	      fpO[ii][y*iWidth + x] = rad[ref_t][ii][y*iWidth + x] > 0 ? (float) rad[ref_t][ii][y*iWidth + x] : 0.;  		    	      
	    }	    			  
	  }

	} /* end for x */

      progress++;
      if ( (progress % 100) == 0 ) {
	int progress_porcentage = progress * 100 / iHeight;
	printf("Status: %d \n",progress_porcentage);
	fflush(stdout);
      }
    } /* end for y */
  
  }							
}	

