/*----------------------------------------------------------------------------

  NLHDR - Non local high dynamic range image generation

  Copyright (c) 2013 cecilia aguerrebere <caguerrebere@gmail.com>
   
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Affero General Public License for more details.

  This code is partially based on the implementation by A. Buades
  
  A. Buades, B. Coll, and J.-M. Morel. Non-Local Means Denoising. 
  Image Processing On Line, 2011

  ----------------------------------------------------------------------------*/
  

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "io_exr.h"
#include "io_pgm.h"
#include "nlhdr.h"

typedef struct
{
  int win;
  int bloc;
  float ih;
  int ref_t;
  int iFrames;		
  float* tau;	
  float* g;
  float muR;
  float sigma2R; 	 
  float iVsat;
  int trans_est;
  int* trans_x;
  int* trans_y;
  const char* prnu;
  const char* output;
  char** input;

} program_argums;


static void usage(const char* name)
{
  printf("\n");
  printf("Non local high dynamic range image generation.\n");
  printf("Version 1.0 - April 2013\n");
  printf("Author - Cecilia Aguerrebere <caguerrebere@gmail.com>\n\n");
  printf("------------------------------------------------------------------------------------------------------------------------\n");
  printf("Usage: %s [parameters] [camera setting] [optional] input1.pgm ... inputN.pgm output.exr\n\n",name);
  printf("Inputs are raw images. Processed images such as JPG or other formats are NOT supported.\n");
  printf("Use the command \n\n");
  printf(" > dcraw -4 -D filename\n\n"); 
  printf("to convert your raw images into PGM (downloadable at http://www.cybercom.net/~dcoffin/dcraw/).\n"); 
  printf("Only PGM images are supported (8 and 16 bits). \n\n");
  printf("The order of the input images is assumed to match that of the exposure times.\n\n");   
  printf("The output is written in 32-bit floating-point EXR format.\n\n");    
  printf("------------------------------------------------------------------------------------------------------------------------\n");
  printf("Parameters:\n");
  printf("  -p <number>                     Half patch size (default 1). Patches are of size 2p+1 x 2p+1.\n");
  printf("  -w <number>                     Half search window (default 5). Search window is of size 2w+1 x 2w+1.\n");
  printf("  -h <number>                     Distance threshold parameter (default 1).\n");
  printf("  -r <number>                     Reference frame number, from 0 to N-1, with N = total frames (default 0).\n\n");
  printf("Camera setting:\n");
  printf("  -n <number> t1 t2 ... tN        Number of frames followed by the exposure times in seconds, separated by spaces.\n");
  printf("                                  The exposure times order must correspond to that of the input images, \n");
  printf("                                  i.e. t1 is the exposure time of the first input image.\n");
  printf("  -g g1 g2 g3 g4                  Gain for each channel separated by spaces. Use -g gain_R gain_G1 gain_G2 gain_B\n");
  printf("                                  if the input images are the RGGB Bayer pattern. Sort the gains accordingly for   \n");
  printf("                                  other Bayer configurations. \n");
  printf("  -m <number>                     Camera readout mean.\n");
  printf("  -v <number>                     Camera readout variance.\n");	
  printf("  -s <number>                     Camera saturation threshold.\n");
  printf("  -u <file name>                  Specify PRNU factors through an 32-bit floating-point EXR image. The image must be the\n");
  printf("                                  same size as the input images. If the -u option is not specified, the PRNU factors are\n");
  printf("                                  set to 1 for all pixels.\n");
  printf("Optional:\n");
  printf("Use to avoid the translation estimation step if the translation values between the images and the reference \n");
  printf("frame are known.\n\n");
  printf("  -t <number> tv1 th1 .. tvN thN  Number of frames followed by the translation values between each frame and the \n");
  printf("                                  reference frame. Give the vertical (rows) and horizontal (columns) translations \n");
  printf("                                  for each frame with respect to the reference frame as a list separated by spaces.\n");
  printf("                                  E.g. If the reference frame is the second frame: -t 4 tv1 th1 0 0 tv3 th3 tv4 th4.\n");
  printf("                                  The translation of the reference frame with itself is 0 0.\n");	
  printf("                                  The translation values order must correspond to that of the input images,\n");
  printf("                                  i.e. tv1 th1 are the vertical and horizontal translations of the first input image\n");
  printf("                                  with respect to the reference frame.\n");	
  printf("\n");
}

static void parse_arguments(program_argums *param, int argc, char *argv[])
{
  char *OptionString = NULL;
  char OptionChar;
  int i;


  if(argc < 5)
    {
      usage(argv[0]);
      exit(EXIT_SUCCESS);
    }


  /* loop to read parameters*/
  for(i = 1; i < argc;)
    {
		
      if(argv[i] && argv[i][0] == '-')
        {
	  if((OptionChar = argv[i][1]) == 0)
            {
	      error("Invalid parameter format.\n");
            }

	  if(argv[i][2])
	    OptionString = &argv[i][2];
	  else if(++i < argc)
	    OptionString = argv[i];
	  else
            {
	      error("Invalid parameter format.\n");
            }

	  switch(OptionChar)
            {
	    case 'p':
	      param->win = atoi(OptionString);
	      if(param->win < 1)
		{
		  error("Invalid half patch size value.\n");
		}
	      break;

	    case 'w':
	      param->bloc = atoi(OptionString);
	      if(param->bloc < 1)
		{
		  error("Invalid half research window size.\n");
		}
	      break;

	    case 'h':
	      param->ih = atof(OptionString);                    
	      if(param->ih < 0)
		{
		  error("Invalid h parameter.\n");
		}
	      break;

	    case 'r':
	      param->ref_t = atoi(OptionString);
	      if(param->ref_t < 0)
		{
		  error("Invalid reference frame.\n");
		}
	      break;

	    case 'n':
	      param->iFrames = atoi(OptionString);
	      if(param->iFrames < 1)
		{
		  error("Invalid number of input frames.\n");
		} else { /* Read the exposure times of each frame */
		param->tau = (float*) malloc(sizeof(float)*param->iFrames);
		int j;
		for (j=0; j < param->iFrames; j++) {
		  i++;
		  /* Check that is a number */
		  if ( isalpha(argv[i][0])  || isalpha(argv[i][1]) )
		    error("Invalid exposure time. Be sure to enter all the exposure times after -n.\n");
		  else if ( atof(argv[i]) > 0 )
		    param->tau[j] = atof(argv[i]);					      
		  else 
		    error("Exposure times must be positive.");
		}	
	      }
	      break;

	    case 'g':
	      param->g = (float*) malloc(sizeof(float)*4);
	      int j;
	      for (j=0; j<4; j++) {
		/* Check that is a number */
		if ( isalpha(argv[i][0])  || isalpha(argv[i][1]) )
		  error("Invalid gain value. Be sure to enter the 4 gain values after -g.\n");
		else if ( atof(argv[i]) > 0 ) 
		  param->g[j] = atof(argv[i]);						      
		else 
		  error("Gain factors must be positive.");
		i++;  
	      }
	      i = i - 1;                    
	      break;

	    case 'm':
	      param->muR = atof(OptionString);
	      if(param->muR < 0)
		{
		  error("Invalid readout mean value.\n");
		}
	      break;

	    case 'v':
	      param->sigma2R = atof(OptionString);
	      if(param->sigma2R < 0)
		{
		  error("Invalid readout variance value.\n");
		}
	      break;

	    case 's':
	      param->iVsat = atof(OptionString);
	      if(param->iVsat < 0)
		{
		  error("Invalid saturation threshold.\n");
		}
	      break;

	    case 'u':
	      param->prnu = argv[i];
	      break;

	    case 't':
	      param->trans_est = atoi(OptionString);
	      if(param->trans_est < 0)
		{						
		  error("Invalid number of frames for global translation estimation.\n");
		} else { /* Read each frame translation with respect to the reference frame */					
		param->trans_y = (int*) malloc(sizeof(int)*param->trans_est);
		param->trans_x = (int*) malloc(sizeof(int)*param->trans_est);
		int j;
		/* Read trans_y row (vertical) and trans_x column (horizontal) translations */
		for (j=0; j<(param->trans_est); j++) {
		  i++;
		  if ( isalpha(argv[i][0]) || isalpha(argv[i][1]) )
		    error("Invalid translation value. Be sure to enter all the translation values after -t.\n");
		  else
		    param->trans_y[j] = atoi(argv[i]);
		  i++;
		  if ( isalpha(argv[i][0]) || isalpha(argv[i][1]) )
		    error("Invalid translation value. Be sure to enter all the translation values after -t.\n");
		  else
		    param->trans_x[j] = atoi(argv[i]);						      
		}
	      }
	      break;

	    case '-':
	      usage(argv[0]);
	      exit(EXIT_FAILURE);

	    default:
	      if(isprint(OptionChar))
		{
		  fprintf(stderr, "Unknown option \"-%c\".\n",
			  OptionChar);
		  exit(EXIT_FAILURE);
		} else
		error("Unknown option.\n");
            }

        }
      else
        {
	  if(!param->input) {
	    param->input = (char**) malloc(sizeof(char*)*param->iFrames);
	    int j;
	    for (j=0; j<param->iFrames; j++) {				 
	      param->input[j] = argv[i];       
	      i++;     	           
	    }
	    i = i - 1;	    
	  }
	  else {
	    if ( !param->output )	      
	      param->output = argv[i];
	  }
        }

      i++;
    }

  if( !param->input[0] )
    {		
      usage(argv[0]);
      exit(EXIT_FAILURE);
    }


  /* If parameters weren't set, set deafult parameters*/
  param->win = param->win > 1 ? param->win : 1;
  param->bloc = param->bloc > 1 ? param->bloc : 5;
  param->ih = param->ih > 0 ? param->ih : 1;
  param->ref_t = param->ref_t>0 ? param->ref_t : 0;
  param->trans_est = param->trans_est>0 ? param->trans_est : 0;
  param->output = param->output ? param->output : "hdr_output.exr";
  param->prnu = param->prnu ? param->prnu : "0";
    
}



int main(int argc, char **argv) {


  /*Initialize the structure param->* to -1 or null */
  program_argums param = {-1, -1, -1, -1, -1, NULL, NULL, -1, -1, -1, -1, NULL, NULL, NULL, NULL, NULL};

  /*Parse command-line arguments*/
  parse_arguments(&param,argc,argv);

  int win = param.win;
  int bloc = param.bloc;
  float ih = param.ih;
  int ref_t = param.ref_t;
  int d_c = 4;
  int iFrames = param.iFrames;	
  float muR = param.muR;
  float sigma2R = param.sigma2R;
  float iVsat = param.iVsat;
  float* tau = param.tau;
  float* g = param.g;
  int i, ii;
	
  /*-----------------------------------------*/		
  /* Load input images */
  float *d_v;	    
  float **fpI[iFrames];
  int d_w = 0, d_h = 0, d_wh = 0;
  int d_w_pgm, d_h_pgm;    	    
    	
  for (i=0; i < iFrames; i++) {			
    d_v = read_pgm_float(param.input[i], &d_w_pgm, &d_h_pgm);				
    if (!d_v) {
      printf("Error: %s not found  or not a correct pgm image. \n", argv[i]);
      exit(-1);
    }

        
    d_w = (d_w_pgm/2);
    d_h = (d_h_pgm/2); 
    d_wh = d_w * d_h;
	    
    fpI[i] = (float**) malloc(d_c*sizeof(float*));		    

    for (ii=0; ii < d_c; ii++) {
      fpI[i][ii] = (float*) malloc(d_wh*sizeof(float)); 
    }
    separate_channels( d_v, fpI[i], d_w_pgm, d_h_pgm );
									
    free(d_v);
  }
  /*-----------------------------------------*/	
  /* Load PRNU factors */	
  float *prnu[d_c];
  float *prnu_file;
  int w_prnu, h_prnu;	
	
  if ( strcmp(param.prnu,"0") == 0 )  { /* A prnu file is not specified. PRNU factors set to 1 */
    prnu_file = (float*) malloc(sizeof(float)*d_w_pgm*d_h_pgm);
    for (i=0; i<d_w_pgm*d_h_pgm; i++)
      prnu_file[i] = 1.0;		
    w_prnu = d_w_pgm;
    h_prnu = d_h_pgm;
  } else {                    /* A prnu file is specified. */
    prnu_file = readEXR_float(param.prnu, &w_prnu, &h_prnu);		
    /* Check that the prnu image and the input images have the same size */
    if ( w_prnu != d_w_pgm || h_prnu != d_h_pgm )
      error("The PRNU image and the input images must have the same size.");
  }
	
  for (ii=0; ii < d_c; ii++) 
    prnu[ii] = (float*) malloc(d_wh*sizeof(float)); 
  separate_channels( prnu_file, prnu, w_prnu, h_prnu );	
  free(prnu_file);				
  /*-----------------------------------------*/
  /* Allocate memory for the output image */
  float *fpO[d_c];	
  for (ii=0; ii < d_c; ii++) {
    fpO[ii] = (float*) malloc(d_wh*sizeof(float));
	
  }
			
  /*--------------------------- Preprocessing ---------------------------*/
   
  /* Global alignment for correct search window centering */		
  int* trans_y = (int*) malloc(sizeof(int)*iFrames);       /* Vector of estimated vertical translations (in rows) of 
							      each image with respect to the reference image */
  int* trans_x = (int*) malloc(sizeof(int)*iFrames);       /* Vector of estimated horizontal translations (in columns) 
							      of each image with respect to the reference image */

  if ( iFrames == 1 ) { /* If there is only one frame no alignment is needed  */						   
    trans_y[0] = 0;
    trans_x[0] = 0;
  } else if ( param.trans_est > 0 ) {/* If the translation vectors are 
                                        provided no alignment is needed  */
    for (i=0; i<param.trans_est; i++) {  /* Read the translation paramters from the input parameters */
      trans_y[i] = param.trans_y[i];
      trans_x[i] = param.trans_x[i];
    }
  } else { /* Perform global translation estimation */
    printf("-------------------------------------------\n");    
    printf("- Starting global translation estimation\n");
    printf("\n");
       
    /* Maximum and minimum tested translations in pixels */
    int y_min = -100;           
    int y_max = 100; 
    int x_min = -100; 
    int x_max = 100;   
       
    global_transl_estimations( fpI, iFrames, d_c, ref_t, d_w*d_h, d_w, d_h, prnu, muR, g, iVsat, tau, y_min, y_max, x_min, x_max, trans_y, trans_x);	   
	   
    printf("\n"); 
    printf("- End of global translation estimation\n\n");  
	   
  }
  /*-------------------------End Preprocessing ---------------------------*/

  print_info ( win, bloc, ih, ref_t, iFrames, tau, d_c, g, muR, sigma2R, iVsat, param.input, trans_y, trans_x, param.prnu );
	
  if ( iFrames > 1 && param.trans_est == 0 ) {
    printf("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
    printf("!!                                                 !!\n"); 
    printf("!! The next run with the same images and reference !!\n");
    printf("!! frame you can skip the translation estimation   !!\n");
    printf("!! step adding the following input option:         !!\n");
    printf("!!                                                 !!\n");
    printf("!! -t %d ",iFrames);
    for(i=0; i<iFrames;i++) {
      printf("%d ",trans_y[i]);
      printf("%d ",trans_x[i]);
    }
    printf("\n");  
    printf("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
  }
	
  /* Non local HDR estimation */
  printf("Starting non local HDR image generation...\n");
  nlhdr(win, bloc, fpI, fpO, iFrames, d_c, d_w, d_h, tau, ref_t, ih, trans_y, trans_x, muR, sigma2R, g, prnu, iVsat);	
  printf("Done.\n\n");
	
  float *fpO_toSave = (float *) malloc(2*d_w*2*d_h*d_c*sizeof(float));
		
  remake_bayer( fpO, fpO_toSave, d_w, d_h );
  writeEXR_float(param.output, fpO_toSave, 2*d_w, 2*d_h);  

  return 0;
}

